# -*- coding: utf-8 -*-

"""
Copyright (C) 2010 Dariusz Suchojad <dsuch at zato.io>

Licensed under LGPLv3, see LICENSE.txt for terms and conditions.
"""

# A place for storing all the defaults values.

web_admin_host = '0.0.0.0'
web_admin_port = 8183

http_plain_server_port = 17010
