# -*- coding: utf-8 -*-

"""
Copyright (C) 2010 Dariusz Suchojad <dsuch at zato.io>

Licensed under LGPLv3, see LICENSE.txt for terms and conditions.
"""

from pkgutil import extend_path
__path__ = extend_path(__path__, __name__)

__import__('pkg_resources').declare_namespace(__name__)
